#ifdef _WIN32
#include "windows.h"
#include "direct.h"
#endif

#include "string.h"

#include "SDL.h"
#include "SDL_ttf.h"
#include "SDL_net.h"
#include "SDL_mixer.h"
#include "SDL_sound.h"
#include "BList.h"
#include "auxiliar.h"
#include "MOG2_auxiliar.h"
#include "music.h"

#include "conditionscript.h"
#include "actionscript.h"
#include "animscript.h"
#include "variables.h"
#include "mog2.h"
#include "object.h"
#include "interfaces.h"
#include "activeobject.h"
#include "activeroom.h"
#include "activeMAP.h"
#include "activemog2.h"
#include "activeinterface.h"
#include "activemog2application.h"
#include "mog2multiplayer.h"

#include "BWindow.h"
#include "BFileListWnd.h"
#include "BActiveWindow.h"
#include "GamePreviewDlg.h"
#include "GameWnd.h"

#ifdef KITSCHY_DEBUG_MEMORY 
#include "debug_memorymanager.h"
#endif

extern int SCREEN_X;
extern int SCREEN_Y;
extern int COLOUR_DEPTH;
extern bool SOUND_ENABLED;
extern bool fullscreen;
extern int extra_flags;
extern char *application_name;

/* Redrawing constant: */ 
int GAME_REDRAWING_PERIOD=40;
extern SDL_Surface *screen_sfc;

extern int buffer_fps,act_fps,fps_timer;
extern int buffer_fps_d,fps_d;


GameWindow::GameWindow(int dx,int dy,TTF_Font *font,MOG2 *g,int gm,bool show_debugmsgp,bool show_fpsp,char *file,List<NWMOG2PlayerInfo> *players) : BActiveWindow(0,0,dx,dy,font)
{
	int i;
	SDL_Rect vp;
	ActiveOBJECT *o;
	List<NWMOG2PlayerInfo> l;
	NWMOG2PlayerInfo *pc=0;

	game=0;
	keyboardstate=0;
	n_LVKs=0;
	game_mode=gm;
	replay_fp=0;
	event_time=-1;
	event_key=0;
	event_val=0;

	show_debugmsg=show_debugmsgp;
	show_fps=show_fpsp;

	app=new ActiveMOG2Application(g);
	game=app->active_game;

	if (game_mode==0 || game_mode==2) {

		l.Instance(*players);
		l.Rewind();
		while(l.Iterate(pc)) {
			vp=pc->vp_area;
			if (pc->start_interface==0) {
				OBJECT *gwo=0;
				MOG2_INTERFACE *i;
				i=new MOG2_INTERFACE(vp.w,vp.h,&gwo);
				interfaces.Add(i);
				objects.Add(gwo);
				o=game->add_player(i,pc->start_room,pc->avatar->name,pc->name,&vp,pc->vp_zoom,0);
				game->viewports[game->viewports.Length()-1]->current_interface->objects[0][0]->gw_focusing.Add(o);
			} else {
				o=game->add_player(pc->start_interface,pc->start_room,pc->avatar->name,pc->name,&vp,pc->vp_zoom,0);
			} /* if */ 
			o->left_key=pc->left_key;
			o->right_key=pc->right_key;
			o->up_key=pc->up_key;
			o->down_key=pc->down_key;
			o->fire1_key=pc->fire1_key;
			o->fire2_key=pc->fire2_key;
			o->fire3_key=pc->fire3_key;
			o->fire4_key=pc->fire4_key;
		} /* while */ 
	} /* if */ 

	switch(game_mode) {
	case 0:
		keyboardstate=new KEYBOARDSTATE *[1];
		keyboardstate[0]=new KEYBOARDSTATE();
		n_LVKs=1;
		break;
	case 1:
		{
			int np,j;
			unsigned int ID_high,ID_low;
			char *name,*name2,*name3,*name4;
			char tmp[80];
			int rx,ry;
			SDL_Rect vp;
			float zoom;
			MAP *start_map;
			ROOM *start_room;
			MOG2_INTERFACE *start_interface;

			replay_fp=fopen(file,"r");

			if (2!=fscanf(replay_fp,"%u %u",&ID_high,&ID_low) ||
				ID_high!=g->ID_high ||
				ID_low!=g->ID_low) return;

			/* read player configuration: */ 
			if (2!=fscanf(replay_fp,"%i %i",&np,&n_LVKs)) return;

			keyboardstate=new KEYBOARDSTATE *[n_LVKs];
			for(j=0;j<n_LVKs;j++) {
				keyboardstate[j]=new KEYBOARDSTATE();
			} /* for */ 

			for(i=0;i<np;i++) {
				int x,y,w,h;
				if (!load_string(replay_fp,&name)) return;
				if (!load_string(replay_fp,&name2)) {
					delete []name;
					return;
				} /* if */ 
				if (2!=fscanf(replay_fp,"%i %i",&rx,&ry)) {
					delete []name;
					delete []name2;
					return;
				} /* if */ 
				if (!load_string(replay_fp,&name3)) {
					delete []name;
					delete []name2;
					return;
				} /* if */ 
				if (5!=fscanf(replay_fp,"%i %i %i %i %f",&x,&y,&w,&h,&zoom)) {
					delete []name;
					delete []name2;
					delete []name3;
					return;
				} /* if */ 
				vp.x=short(x);
				vp.y=short(y);
				vp.w=short(w);
				vp.h=short(h);

				if (1!=fscanf(replay_fp,"%s",tmp)) {
					delete []name;
					delete []name2;
					delete []name3;
					return;
				} /* if */ 
				if (strcmp(tmp,"INTERFACE")==0) {
					if (!load_string(replay_fp,&name4)) {
						delete []name;
						delete []name2;
						delete []name3;
						return;
					} /* if */ 
				} else {
					name4=0;
				} /* if */ 

				/* Search map: */ 
				{
					MAP *m=0;
					List<MAP> l;
					ROOM *r=0;
					List<ROOM> l2;
					MOG2_INTERFACE *interf=0;
					List<MOG2_INTERFACE> l3;
					bool found=false;

					start_map=0;

					l.Instance(g->maps);
					l.Rewind();
					while(!found && l.Iterate(m)) {
						if (m->cmp(name3)) {
							start_map=m;
							found=true;
						} /* if */ 
					} /* while */ 

					if (!found) {
						delete []name;
						delete []name2;
						delete []name3;
						delete []name4;
						return;
					} /* if */ 

					/* Search room: */ 
					found=false;

					start_room=0;
					
					l2.Instance(start_map->rooms);
					l2.Rewind();
					while(!found && l2.Iterate(r)) {
						if (r->get_x()==rx &&
							r->get_y()==ry) {
							start_room=r;
							found=true;
						} /* if */ 
					} /* while */ 

					/* Search interface: */ 
					start_interface=0;
					if (name4!=0) {
						found=false;
						l3.Instance(g->interfaces);
						l3.Rewind();
						while(!found && l3.Iterate(interf)) {
							if (interf->cmp(name4)) {
								start_interface=interf;
								found=true;
							} /* if */ 
						} /* while */ 
					} /* if */ 

					if (!found) {
						delete []name;
						delete []name2;
						delete []name3;
						delete []name4;
						return;
					} /* if */ 
					if (pc->start_interface==0) {
						OBJECT *gwo=0;
						MOG2_INTERFACE *i;
						i=new MOG2_INTERFACE(vp.w,vp.h,&gwo);
						interfaces.Add(i);
						objects.Add(gwo);
						o=game->add_player(i,start_room,name,name2,&vp,zoom,0);
						game->viewports[game->viewports.Length()-1]->current_interface->objects[0][0]->gw_focusing.Add(o);
					} else {
						o=game->add_player(start_interface,start_room,name,name2,&vp,zoom,0);
					} /* if */ 

				}

				if (8!=fscanf(replay_fp,"%i %i %i %i %i %i %i %i",
							  &o->left_key,&o->right_key,&o->up_key,&o->down_key,
							  &o->fire1_key,&o->fire2_key,&o->fire3_key,&o->fire4_key)) {
					delete []name;
					delete []name2;
					delete []name3;
					delete []name4;
					return;
				} /* if */ 

				delete []name;
				delete []name2;
				delete []name3;
				delete []name4;
			} /* for */ 

		}
		break;
	case 2:
		{
			int i;
			NWMOG2PlayerInfo *pi;
			char *tmp;

			keyboardstate=new KEYBOARDSTATE *[1];
			keyboardstate[0]=new KEYBOARDSTATE();
			n_LVKs=1;

			replay_fp=fopen(file,"w");
			fprintf(replay_fp,"%u %u\n",g->ID_high,g->ID_low);
			fprintf(replay_fp,"%i %i\n",players->Length(),1);

			for(i=0;i<players->Length();i++) {	
				pi=(*players)[i];
				encode_string(pi->avatar->name,&tmp);
				fprintf(replay_fp,"%s\n",tmp);
				delete tmp;

				encode_string(pi->name,&tmp);
				fprintf(replay_fp,"%s\n",tmp);
				delete tmp;

				fprintf(replay_fp,"%i %i\n",pi->start_room->get_x(),pi->start_room->get_y());

				encode_string(pi->start_map->name,&tmp);
				fprintf(replay_fp,"%s\n",tmp);
				delete tmp;

				fprintf(replay_fp,"%i %i %i %i %f\n",pi->vp_area.x,pi->vp_area.y,
												 	 pi->vp_area.w,pi->vp_area.h,
													 pi->vp_zoom);

				if (pi->start_interface!=0) {
					encode_string(pi->start_interface->name,&tmp);
					fprintf(replay_fp,"INTERFACE %s\n",tmp);
					delete tmp;
				} else {
					fprintf(replay_fp,"NOINTERFACE\n");
				} /* if */ 

				fprintf(replay_fp,"%i %i %i %i %i %i %i %i\n",pi->left_key,pi->right_key,pi->up_key,pi->down_key,
															  pi->fire1_key,pi->fire2_key,pi->fire3_key,pi->fire4_key);		
			} /* for */		
		}
		break;
	}/* switch */ 
	
} /* GameWindow::GameWindow */ 


GameWindow::~GameWindow()
{
	int i;

	for(i=0;i<n_LVKs;i++) {
		delete keyboardstate[i];
		keyboardstate[i]=0;
	} /* if */ 
	delete []keyboardstate;
	if (game_mode==2) {
		/* Saving replay: */ 
		fprintf(replay_fp,"END\n");
	} /* if */ 

	if (game_mode!=0) fclose(replay_fp);

	delete app;
} /* GameWindow:: */ 


void GameWindow::Execute(void)
{
	int time,act_time;
	SDL_Event event;
	bool quit=false;

	active_window=this;

	SDL_QuitSubSystem(SDL_INIT_VIDEO);
	SDL_InitSubSystem(SDL_INIT_VIDEO);
	screen_sfc = SDL_SetVideoMode(old_dx, old_dy, COLOUR_DEPTH, (fullscreen ? SDL_FULLSCREEN : 0)|extra_flags);
	SDL_WM_SetCaption(application_name, 0);
	SDL_ShowCursor(SDL_DISABLE);
	SDL_EnableUNICODE(1);

//	if (SOUND_ENABLED) {
//		init_music(game);
//		Mix_HookMusic(myMusicPlayer, (void *)0);
//	} /* if */ 
	
	time=SDL_GetTicks();
	while (!quit) {
		while( SDL_PollEvent( &event ) ) {
			bool captured=false;

            switch( event.type ) {
                /* Keyboard event */ 
                case SDL_KEYDOWN:
					if (event.key.keysym.sym==SDLK_F12 ||
						event.key.keysym.sym==SDLK_ESCAPE) {
						captured=true;
						quit = true;
					} /* if */ 

					if (event.key.keysym.sym==SDLK_RETURN) {
						SDLMod modifiers=SDL_GetModState();

						if ((modifiers&KMOD_ALT)!=0) {
							captured=true;

							/* Toogle FULLSCREEN mode: */ 
							if (fullscreen) fullscreen=false;
									   else fullscreen=true;
							SDL_QuitSubSystem(SDL_INIT_VIDEO);
							SDL_InitSubSystem(SDL_INIT_VIDEO);
							if (SDL_WasInit(SDL_INIT_VIDEO)) {
								screen_sfc = SDL_SetVideoMode(old_dx, old_dy, COLOUR_DEPTH, (fullscreen ? SDL_FULLSCREEN : 0)|extra_flags);
								SDL_WM_SetCaption(application_name, 0);
								SDL_ShowCursor(SDL_DISABLE);
								SDL_EnableUNICODE(1);
							} else {
								quit = true;
							} /* if */ 
						} /* if */ 
					} /* if */ 
                    break;

                /* SDL_QUIT event (window close) */ 
                case SDL_QUIT:
					captured=true;

                    quit = true;
                    break;
            } /* switch */ 

			if (!captured) {
				if (!this->event(&event)) quit=true;
			} /* if */ 
        } /* while */ 

		act_time=SDL_GetTicks();
		if (act_time-time>=GAME_REDRAWING_PERIOD)
		{
			act_fps++;
			if ((act_time-fps_timer)>1000) {
				fps_timer=act_time;
				buffer_fps=act_fps;
				buffer_fps_d=fps_d;
				act_fps=0;
				fps_d=0;
			} /* if */ 

			time+=GAME_REDRAWING_PERIOD;
			if ((act_time-time)>8*GAME_REDRAWING_PERIOD) time=act_time;

			check_mouse();
			if (!cycle(old_x,old_y)) {
				quit=true;
			} /* if */ 
		}

		if (!quit) {
			fps_d++;
			SDL_FillRect(screen_sfc,0,0);
			if (game!=0) {
				game->draw(screen_sfc);
//				game->sound(0,-1,-1);
			} /* if */ 

			/* show INFO: */ 
			if (game!=0) {
				char tmp[256];
				List<ActiveOBJECT> l;
				ActiveOBJECT *ao=0;
				int y=0;

				if (show_fps) {
					sprintf(tmp,"FPS: %i/%i",buffer_fps,buffer_fps_d);
					print_left(tmp,font,255,255,255,screen_sfc,0,0);
				} /* if */ 
				if (show_debugmsg) {
					sprintf(tmp,"ACTIVE MAPS: [ ");
					{
						List<ActiveMAP> l2;
						ActiveMAP *map=0;
						char tmp2[80];

						l2.Instance(game->active_maps);
						l2.Rewind();
						while(l2.Iterate(map)) {
							sprintf(tmp2,"%s ",map->map->name);
							strcat(tmp,tmp2);
						} /* while */ 
						strcat(tmp,"]");
					}
					print_left(tmp,font,255,255,255,screen_sfc,0,10);

					y=20;
					{
						List<ActiveMAP> l4;
						ActiveMAP *map=0;

						l4.Instance(game->active_maps);
						l4.Rewind();
						while(l4.Iterate(map)) {
							sprintf(tmp,"MAP R.: %i  -  INACT R.: %i  -  ACT R.: %i ",map->map->rooms.Length(),map->inactive_rooms.Length(),map->active_rooms.Length());
							{
								List<ActiveROOM> l2;
								List<ROOM> l3;
								ROOM *rtmp=0;
								ActiveROOM *r=0;
								char tmp2[80];
								int i,j;

								l2.Instance(map->active_rooms);
								l2.Rewind();
								while(l2.Iterate(r)) {
									l3.Instance(map->map->rooms);
									l3.Rewind();
									i=-1;
									j=0;
									while(l3.Iterate(rtmp)) {
										j++;
										if (rtmp==r->room) i=j;
									} /* while */ 
									sprintf(tmp2,"%i[ %i,%i ]{%i,%i} ",i,r->objects[0].Length(),r->objects[1].Length(),
																	   r->nonpassive_objects[0].Length(),r->nonpassive_objects[1].Length());
									strcat(tmp,tmp2);
								} /* while */ 
							}
							print_left(tmp,font,255,255,255,screen_sfc,0,y);
							y+=10;
						} /* while */ 
					}

					l.Instance(game->player_objects);
					l.Rewind();
					while(l.Iterate(ao)) {
						int ox,oy;
						int x1,y1,x2,y2;
						ao->get_bbox(&x1,&y1,&x2,&y2);
						ao->hotspot_coordinates(&ox,&oy);
						sprintf(tmp,"CHARACTER, X,Y: %i,%i,%i,%i  -  ENERGY: %i  -  STATE: %i (%i)  -  INSTANCES:  %i ",
							x1,y1,x2,y2,
							ao->energy,ao->state,ao->state_timmer,
							ao->other_instances.Length()+1);
					
						{
							List<ActiveOBJECT> l2;
							ActiveOBJECT *o=0;
							char tmp2[80];

							ao->absolute_hotspot_coordinates(&ox,&oy);
							sprintf(tmp2,"[%i,%i] ",ox,oy);
							strcat(tmp,tmp2);

							l2.Instance(ao->other_instances);
							l2.Rewind();
							while(l2.Iterate(o)) {
								o->absolute_hotspot_coordinates(&ox,&oy);
								sprintf(tmp2,"[%i,%i] ",ox,oy);
								strcat(tmp,tmp2);
							} /* while */ 
						}

						print_left(tmp,font,255,255,255,screen_sfc,0,y);
						y+=10;
					} /* if */ 
				} /* if */ 
			} /* if */ 
			SDL_Flip(screen_sfc);
		} /* if */ 

	}

//	if (SOUND_ENABLED) {
//		Mix_HookMusic(0, 0);
//		init_music(game);
//	} /* if */ 

	SDL_QuitSubSystem(SDL_INIT_VIDEO);
	SDL_InitSubSystem(SDL_INIT_VIDEO);
	screen_sfc = SDL_SetVideoMode(SCREEN_X, SCREEN_Y, COLOUR_DEPTH, (fullscreen ? SDL_FULLSCREEN : 0));
	SDL_WM_SetCaption(application_name, 0);
	SDL_ShowCursor(SDL_DISABLE);
	SDL_EnableUNICODE(1);

} /* GameWindow::Execute */ 


bool GameWindow::cycle(int stx,int sty)
{
	int i;
	bool retval=true;

	dirty=true;

	if (game_mode==0 || game_mode==2) {
		keyboardstate[0]->cycle();

		if (game!=0) {
			retval=game->cycle(keyboardstate,1);
			app->current_cycle++;
		} /* if */ 

		if (game_mode==2) {
			/* Saving replay: */ 
			fprintf(replay_fp,"NEWCYCLE\n");
			for(i=0;i<n_LVKs;i++) keyboardstate[i]->save(replay_fp);
		} /* if */ 

	} /* if */ 

	if (game_mode==1) {
		char tmp[80];

		/* Load replay: */ 
		if (1!=fscanf(replay_fp,"%s",tmp) ||
			strcmp(tmp,"NEWCYCLE")!=0) return false;

		for(i=0;i<n_LVKs;i++) keyboardstate[i]->load(replay_fp);

		if (game!=0) {
			retval=game->cycle(keyboardstate,1);
			app->current_cycle++;
		} /* if */ 
	} /* if */ 

	for(i=0;i<n_LVKs;i++) keyboardstate[i]->keyevents.Delete();

	return retval;
} /* GameWindow::cycle */ 


void GameWindow::contens_size(int *dx,int *dy)
{
} /* GameWindow::contens_size */ 


void GameWindow::draw_contens(SDL_Surface *sfc,int stx,int sty,int dx,int dy)
{
} /* GameWindow::draw_contens */ 


bool GameWindow::event(SDL_Event *event)
{
	if (active_window!=this) return true;

	if (event->type==SDL_KEYDOWN && 
		(game_mode==0 || game_mode==2)) {
		/* Keyboard event */ 
		SDL_keysym *ks;
		ks=new SDL_keysym();
		*ks=event->key.keysym;
		keyboardstate[0]->keyevents.Add(ks);
    } /* if */ 

	return true;
} /* GameWindow::Event */ 
